@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import abc
 import asyncio
 import concurrent.futures
@@ -14,7 +16,7 @@ import traceback
 from concurrent.futures import ThreadPoolExecutor
 from contextlib import suppress
 from pathlib import Path
-from typing import Any, Callable, Coroutine, Dict, List, Optional, Set, Tuple, Union
+from typing import Any, Callable, Coroutine, Literal
 
 import adaptive
 import cloudpickle
@@ -28,10 +30,14 @@ from tinydb import Query, TinyDB
 
 from adaptive_scheduler.scheduler import BaseScheduler
 from adaptive_scheduler.utils import (
+    _DATAFRAME_FORMATS,
     _deserialize,
     _progress,
     _remove_or_move_files,
     _serialize,
+    cloudpickle_learners,
+    fname_to_learner_fname,
+    load_dataframes,
     load_parallel,
     maybe_lst,
 )
@@ -51,9 +57,9 @@ class MaxRestartsReached(Exception):
 
 class _BaseManager(metaclass=abc.ABCMeta):
     def __init__(self) -> None:
-        self.ioloop: Optional[asyncio.events.AbstractEventLoop] = None
-        self._coro: Optional[Coroutine] = None
-        self.task: Optional[asyncio.Task] = None
+        self.ioloop: asyncio.events.AbstractEventLoop | None = None
+        self._coro: Coroutine | None = None
+        self.task: asyncio.Task | None = None
 
     def start(self):
         if self.is_started:
@@ -68,13 +74,12 @@ class _BaseManager(metaclass=abc.ABCMeta):
     def is_started(self) -> bool:
         return self.task is not None
 
-    def cancel(self) -> Optional[bool]:
+    def cancel(self) -> bool | None:
         if self.is_started:
             return self.task.cancel()
 
     def _setup(self):
         """Is run in the beginning of `self.start`."""
-        pass
 
     @abc.abstractmethod
     async def _manage(self) -> None:
@@ -112,8 +117,8 @@ class DatabaseManager(_BaseManager):
         url: str,
         scheduler: BaseScheduler,
         db_fname: str,
-        learners: List[adaptive.BaseLearner],
-        fnames: Union[List[str], List[List[str]]],
+        learners: list[adaptive.BaseLearner],
+        fnames: list[str] | list[list[str]],
         overwrite_db: bool = True,
     ) -> None:
         super().__init__()
@@ -128,16 +133,17 @@ class DatabaseManager(_BaseManager):
             job_id=None, is_done=False, log_fname=None, job_name=None, output_logs=[]
         )
 
-        self._last_reply: Union[str, Exception, None] = None
-        self._last_request: Optional[Tuple[str, ...]] = None
-        self.failed: List[Dict[str, Any]] = []
+        self._last_reply: str | Exception | None = None
+        self._last_request: tuple[str, ...] | None = None
+        self.failed: list[dict[str, Any]] = []
 
     def _setup(self) -> None:
         if os.path.exists(self.db_fname) and not self.overwrite_db:
             return
         self.create_empty_db()
+        cloudpickle_learners(self.learners, self.fnames, with_progress_bar=True)
 
-    def update(self, queue: Optional[Dict[str, Dict[str, str]]] = None) -> None:
+    def update(self, queue: dict[str, dict[str, str]] | None = None) -> None:
         """If the ``job_id`` isn't running anymore, replace it with None."""
         if queue is None:
             queue = self.scheduler.queue(me_only=True)
@@ -167,7 +173,7 @@ class DatabaseManager(_BaseManager):
         with TinyDB(self.db_fname) as db:
             db.insert_multiple(entries)
 
-    def as_dicts(self) -> List[Dict[str, str]]:
+    def as_dicts(self) -> list[dict[str, str]]:
         with TinyDB(self.db_fname) as db:
             return db.all()
 
@@ -178,9 +184,7 @@ class DatabaseManager(_BaseManager):
             f.replace(self.scheduler._JOB_ID_VARIABLE, job_id) for f in output_fnames
         ]
 
-    def _start_request(
-        self, job_id: str, log_fname: str, job_name: str
-    ) -> Optional[str]:
+    def _start_request(self, job_id: str, log_fname: str, job_name: str) -> str | None:
         Entry = Query()
         with TinyDB(self.db_fname) as db:
             if db.contains(Entry.job_id == job_id):
@@ -209,7 +213,7 @@ class DatabaseManager(_BaseManager):
             )
         return entry["fname"]
 
-    def _stop_request(self, fname: Union[str, List[str]]) -> None:
+    def _stop_request(self, fname: str | list[str]) -> None:
         fname = maybe_lst(fname)  # if a BalancingLearner
         Entry = Query()
         with TinyDB(self.db_fname) as db:
@@ -219,7 +223,7 @@ class DatabaseManager(_BaseManager):
             )  # make sure the entry exists
             db.update(reset, Entry.fname == fname)
 
-    def _stop_requests(self, fnames: List[Union[str, List[str]]]) -> None:
+    def _stop_requests(self, fnames: list[str | list[str]]) -> None:
         # Same as `_stop_request` but optimized for processing many `fnames` at once
         fnames = {str(maybe_lst(fname)) for fname in fnames}
         with TinyDB(self.db_fname) as db:
@@ -227,7 +231,7 @@ class DatabaseManager(_BaseManager):
             doc_ids = [e.doc_id for e in db.all() if str(e["fname"]) in fnames]
             db.update(reset, doc_ids=doc_ids)
 
-    def _dispatch(self, request: Tuple[str, ...]) -> Union[str, Exception, None]:
+    def _dispatch(self, request: tuple[str, ...]) -> str | Exception | None:
         request_type, *request_arg = request
         log.debug("got a request", request=request)
         try:
@@ -239,13 +243,8 @@ class DatabaseManager(_BaseManager):
                 fname = self._start_request(**kwargs)
                 if fname is None:
                     raise RuntimeError("No more learners to run in the database.")
-                learner = next(
-                    learner
-                    for learner, f in zip(self.learners, self.fnames)
-                    if maybe_lst(f) == fname
-                )
                 log.debug("choose a fname", fname=fname, **kwargs)
-                return learner, fname
+                return fname
             elif request_type == "stop":
                 fname = request_arg[0]  # workers send us the fname they were given
                 log.debug("got a stop request", fname=fname)
@@ -324,7 +323,7 @@ class JobManager(_BaseManager):
 
     def __init__(
         self,
-        job_names: List[str],
+        job_names: list[str],
         database_manager: DatabaseManager,
         scheduler: BaseScheduler,
         interval: int = 30,
@@ -347,7 +346,7 @@ class JobManager(_BaseManager):
         """Equivalent to ``self.max_fails_per_job * len(self.job_names)``"""
         return self.max_fails_per_job * len(self.job_names)
 
-    def _queued(self, queue) -> Set[str]:
+    def _queued(self, queue) -> set[str]:
         return {
             job["job_name"]
             for job in queue.values()
@@ -407,8 +406,8 @@ class JobManager(_BaseManager):
 
 
 def logs_with_string_or_condition(
-    error: Union[str, Callable[[List[str]], bool]], database_manager: DatabaseManager
-) -> List[Tuple[str, List[str]]]:
+    error: str | Callable[[list[str]], bool], database_manager: DatabaseManager
+) -> list[tuple[str, list[str]]]:
     """Get jobs that have `string` (or apply a callable) inside their log-file.
 
     Either use `string` or `error`.
@@ -483,10 +482,10 @@ class KillManager(_BaseManager):
         self,
         scheduler: BaseScheduler,
         database_manager: DatabaseManager,
-        error: Union[str, Callable[[List[str]], bool]] = "srun: error:",
+        error: str | Callable[[list[str]], bool] = "srun: error:",
         interval: int = 600,
         max_cancel_tries: int = 5,
-        move_to: Optional[str] = None,
+        move_to: str | None = None,
     ):
         super().__init__()
         self.scheduler = scheduler
@@ -496,8 +495,8 @@ class KillManager(_BaseManager):
         self.max_cancel_tries = max_cancel_tries
         self.move_to = move_to
 
-        self.cancelled: List[str] = []
-        self.deleted: List[str] = []
+        self.cancelled: list[str] = []
+        self.deleted: list[str] = []
 
     async def _manage(self) -> None:
         while True:
@@ -508,8 +507,8 @@ class KillManager(_BaseManager):
                     self.error, self.database_manager
                 )
 
-                to_cancel: List[str] = []
-                to_delete: List[str] = []
+                to_cancel: list[str] = []
+                to_delete: list[str] = []
                 for job_name, fnames in failed_jobs:
                     to_cancel.append(job_name)
                     to_delete.extend(fnames)
@@ -548,10 +547,15 @@ def _make_default_run_script(
     url: str,
     save_interval: int,
     log_interval: int,
-    goal: Union[Callable[[adaptive.BaseLearner], bool], None] = None,
-    runner_kwargs: Optional[Dict[str, Any]] = None,
+    goal: Callable[[adaptive.BaseLearner], bool] | None = None,
+    runner_kwargs: dict[str, Any] | None = None,
     run_script_fname: str = "run_learner.py",
     executor_type: str = "mpi4py",
+    loky_start_method: Literal[
+        "loky", "loky_int_main", "spawn", "fork", "forkserver"
+    ] = "loky",
+    save_dataframe: bool = False,
+    dataframe_format: _DATAFRAME_FORMATS = "parquet",
 ) -> None:
     default_runner_kwargs = dict(shutdown_executor=True)
     runner_kwargs = dict(default_runner_kwargs, goal=goal, **(runner_kwargs or {}))
@@ -579,14 +583,17 @@ def _make_default_run_script(
         serialized_runner_kwargs=serialized_runner_kwargs,
         save_interval=save_interval,
         log_interval=log_interval,
+        loky_start_method=loky_start_method,
+        save_dataframe=save_dataframe,
+        dataframe_format=dataframe_format,
     )
 
     with open(run_script_fname, "w", encoding="utf-8") as f:
         f.write(template)
 
 
-def _get_infos(fname: str, only_last: bool = True) -> List[str]:
-    status_lines: List[str] = []
+def _get_infos(fname: str, only_last: bool = True) -> list[str]:
+    status_lines: list[str] = []
     with open(fname, encoding="utf-8") as f:
         lines = f.readlines()
         for line in reversed(lines):
@@ -600,7 +607,7 @@ def _get_infos(fname: str, only_last: bool = True) -> List[str]:
 
 
 def parse_log_files(
-    job_names: List[str],
+    job_names: list[str],
     database_manager: DatabaseManager,
     scheduler,
     only_last: bool = True,
@@ -653,7 +660,7 @@ def parse_log_files(
     return pd.DataFrame(infos)
 
 
-def _get_all_files(job_names: List[str], scheduler: BaseScheduler) -> List[str]:
+def _get_all_files(job_names: list[str], scheduler: BaseScheduler) -> list[str]:
     log_fnames = [scheduler.log_fname(name) for name in job_names]
     output_fnames = [scheduler.output_fnames(name) for name in job_names]
     output_fnames = sum(output_fnames, [])
@@ -670,11 +677,11 @@ def _get_all_files(job_names: List[str], scheduler: BaseScheduler) -> List[str]:
     return all_files
 
 
-def cleanup(
-    job_names: List[str],
+def cleanup_scheduler_files(
+    job_names: list[str],
     scheduler: BaseScheduler,
     with_progress_bar: bool = True,
-    move_to: Optional[str] = None,
+    move_to: str | None = None,
 ) -> None:
     """Cleanup the scheduler log-files files.
 
@@ -777,6 +784,16 @@ class RunManager(_BaseManager):
         Keyword arguments for the `JobManager` function that aren't set in ``__init__`` here.
     kill_manager_kwargs : dict, default: None
         Keyword arguments for the `KillManager` function that aren't set in ``__init__`` here.
+    loky_start_method : str
+        Loky start method, by default "loky".
+    cleanup_first : bool
+        Cancel all previous jobs generated by the same RunManager and clean logfiles.
+    save_dataframe : bool
+        Whether to periodically save the learner's data as a `pandas.DataFame`.
+    dataframe_format : str
+        The format in which to save the `pandas.DataFame`. See the type hint for the options.
+    max_log_lines : int
+        The maximum number of lines to display in the log viewer widget.
 
     Attributes
     ----------
@@ -825,25 +842,36 @@ class RunManager(_BaseManager):
     def __init__(
         self,
         scheduler: BaseScheduler,
-        learners: List[adaptive.BaseLearner],
-        fnames: List[str],
-        goal: Union[Callable[[adaptive.BaseLearner], bool], None] = None,
+        learners: list[adaptive.BaseLearner],
+        fnames: list[str],
+        goal: Callable[[adaptive.BaseLearner], bool] | None = None,
         check_goal_on_start: bool = True,
-        runner_kwargs: Optional[dict] = None,
-        url: Optional[str] = None,
+        runner_kwargs: dict | None = None,
+        url: str | None = None,
         save_interval: int = 300,
         log_interval: int = 300,
         job_name: str = "adaptive-scheduler",
         job_manager_interval: int = 60,
         kill_interval: int = 60,
-        kill_on_error: Union[str, Callable[[List[str]], bool], None] = "srun: error:",
-        move_old_logs_to: Optional[str] = "old_logs",
-        db_fname: Optional[str] = None,
+        kill_on_error: str | Callable[[list[str]], bool] | None = "srun: error:",
+        move_old_logs_to: str | None = "old_logs",
+        db_fname: str | None = None,
         overwrite_db: bool = True,
-        job_manager_kwargs: Optional[Dict[str, Any]] = None,
-        kill_manager_kwargs: Optional[Dict[str, Any]] = None,
+        job_manager_kwargs: dict[str, Any] | None = None,
+        kill_manager_kwargs: dict[str, Any] | None = None,
+        loky_start_method: Literal[
+            "loky", "loky_int_main", "spawn", "fork", "forkserver"
+        ] = "loky",
+        cleanup_first: bool = False,
+        save_dataframe: bool = False,
+        # TODO: use _DATAFRAME_FORMATS instead of literal in ≥Python 3.10
+        dataframe_format: Literal[
+            "parquet", "csv", "hdf", "pickle", "feather", "excel", "json"
+        ] = "parquet",
+        max_log_lines: int = 500,
     ):
         super().__init__()
+
         # Set from arguments
         self.scheduler = scheduler
         self.goal = goal
@@ -860,10 +888,25 @@ class RunManager(_BaseManager):
         self.overwrite_db = overwrite_db
         self.job_manager_kwargs = job_manager_kwargs or {}
         self.kill_manager_kwargs = kill_manager_kwargs or {}
+        self.loky_start_method = loky_start_method
+        self.save_dataframe = save_dataframe
+        self.dataframe_format = dataframe_format
+        self.max_log_lines = max_log_lines
+
+        if self.save_dataframe:
+            import pkg_resources
+
+            required = pkg_resources.parse_version("0.14.0")
+            current = pkg_resources.parse_version(adaptive.__version__)
+            if current < required:
+                raise RuntimeError(
+                    "`save_dataframe` requires adaptive version "
+                    f"of at least {required}, currently using {current}."
+                )
 
         # Set in methods
-        self.start_time: Optional[float] = None
-        self.end_time: Optional[float] = None
+        self.start_time: float | None = None
+        self.end_time: float | None = None
 
         # Set on init
         self.learners = learners
@@ -876,6 +919,11 @@ class RunManager(_BaseManager):
             assert isinstance(self.fnames[0], str)
 
         self.job_names = [f"{self.job_name}-{i}" for i in range(len(self.learners))]
+
+        if cleanup_first:
+            self.scheduler.cancel(self.job_names)
+            self.cleanup(True)
+
         self.url = url or get_allowed_url()
         self.database_manager = DatabaseManager(
             self.url,
@@ -914,6 +962,9 @@ class RunManager(_BaseManager):
             self.runner_kwargs,
             self.scheduler.run_script,
             self.scheduler.executor_type,
+            self.loky_start_method,
+            self.save_dataframe,
+            self.dataframe_format,
         )
         self.database_manager.start()
         if self.check_goal_on_start:
@@ -954,9 +1005,14 @@ class RunManager(_BaseManager):
             if self.status() != "running":
                 os.remove(self.scheduler.run_script)
 
+        for fname in self.fnames:
+            fname_cloudpickle = fname_to_learner_fname(fname)
+            with suppress(FileNotFoundError):
+                os.remove(fname_cloudpickle)
+
         _delete_old_ipython_profiles(self.scheduler)
 
-        cleanup(
+        cleanup_scheduler_files(
             job_names=self.job_names,
             scheduler=self.scheduler,
             with_progress_bar=True,
@@ -994,7 +1050,7 @@ class RunManager(_BaseManager):
         if self.task is not None:
             self.task.print_stack()
 
-    def get_database(self) -> List[Dict[str, Any]]:
+    def get_database(self) -> list[dict[str, Any]]:
         """Get the database as a `pandas.DataFrame`."""
         return pd.DataFrame(self.database_manager.as_dicts())
 
@@ -1057,6 +1113,13 @@ class RunManager(_BaseManager):
     def info(self):
         return info(self)
 
+    def load_dataframes(self):
+        """Load the `pandas.DataFrame`s with the most recently saved learners data."""
+        if not self.save_dataframe:
+            raise ValueError("The `save_dataframe` option was not set to True.")
+        df = load_dataframes(self.fnames, format=self.dataframe_format)
+        return df
+
 
 def periodically_clean_ipython_profiles(scheduler, interval: int = 600):
     """Periodically remove old IPython profiles.
