@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import abc
 import collections.abc
 import functools
@@ -15,11 +17,12 @@ from contextlib import suppress
 from inspect import signature
 from multiprocessing import Manager
 from pathlib import Path
-from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
+from typing import Any, Callable, Literal, Sequence
 
 import adaptive
 import cloudpickle
 import numpy as np
+import pandas as pd
 import toolz
 from adaptive.notebook_integration import in_ipynb
 from ipyparallel import Client
@@ -79,11 +82,11 @@ def split(seq: collections.abc.Iterable, n_parts: int):
 
 
 def split_in_balancing_learners(
-    learners: List[adaptive.BaseLearner],
-    fnames: List[str],
+    learners: list[adaptive.BaseLearner],
+    fnames: list[str],
     n_parts: int,
     strategy: str = "npoints",
-) -> Tuple[List[adaptive.BaseLearner], List[str]]:
+) -> tuple[list[adaptive.BaseLearner], list[str]]:
     r"""Split a list of learners and fnames into `adaptive.BalancingLearner`\s.
 
     Parameters
@@ -112,8 +115,8 @@ def split_in_balancing_learners(
 
 
 def split_sequence_learner(
-    big_learner, n_learners: int, folder: Union[str, Path] = ""
-) -> Tuple[List[adaptive.SequenceLearner], List[str]]:
+    big_learner, n_learners: int, folder: str | Path = ""
+) -> tuple[list[adaptive.SequenceLearner], list[str]]:
     r"""Split a sinlge `~adaptive.SequenceLearner` into
     mutiple `adaptive.SequenceLearner`\s (with the data loaded) and fnames.
 
@@ -158,8 +161,8 @@ def split_sequence_in_sequence_learners(
     function: Callable[[Any], Any],
     sequence: Sequence[Any],
     n_learners: int,
-    folder: Union[str, Path] = "",
-) -> Tuple[List[adaptive.SequenceLearner], List[str]]:
+    folder: str | Path = "",
+) -> tuple[list[adaptive.SequenceLearner], list[str]]:
     r"""Split a sequenceinto `adaptive.SequenceLearner`\s and fnames.
 
     Parameters
@@ -193,8 +196,8 @@ def split_sequence_in_sequence_learners(
 
 
 def combine_sequence_learners(
-    learners: List[adaptive.SequenceLearner],
-    big_learner: Optional[adaptive.SequenceLearner] = None,
+    learners: list[adaptive.SequenceLearner],
+    big_learner: adaptive.SequenceLearner | None = None,
 ) -> adaptive.SequenceLearner:
     r"""Combine several `~adaptive.SequenceLearner`\s into a single
     `~adaptive.SequenceLearner` any copy over the data.
@@ -253,7 +256,7 @@ def copy_from_sequence_learner(
             learner_to.tell((i, key), v)
 
 
-def _get_npoints(learner: adaptive.BaseLearner) -> Optional[int]:
+def _get_npoints(learner: adaptive.BaseLearner) -> int | None:
     with suppress(AttributeError):
         return learner.npoints
     with suppress(AttributeError):
@@ -274,7 +277,7 @@ def _progress(
 
 
 def combo_to_fname(
-    combo: Dict[str, Any], folder: Optional[str] = None, ext: Optional[str] = ".pickle"
+    combo: dict[str, Any], folder: str | None = None, ext: str | None = ".pickle"
 ) -> str:
     """Converts a dict into a human readable filename."""
     fname = "__".join(f"{k}_{v}" for k, v in combo.items()) + ext
@@ -284,9 +287,9 @@ def combo_to_fname(
 
 
 def combo2fname(
-    combo: Dict[str, Any],
-    folder: Optional[Union[str, Path]] = None,
-    ext: Optional[str] = ".pickle",
+    combo: dict[str, Any],
+    folder: str | Path | None = None,
+    ext: str | None = ".pickle",
     sig_figs: int = 8,
 ) -> str:
     """Converts a dict into a human readable filename.
@@ -300,10 +303,10 @@ def combo2fname(
 
 
 def add_constant_to_fname(
-    combo: Dict[str, Any],
-    constant: Dict[str, Any],
-    folder: Optional[Union[str, Path]] = None,
-    ext: Optional[str] = ".pickle",
+    combo: dict[str, Any],
+    constant: dict[str, Any],
+    folder: str | Path | None = None,
+    ext: str | None = ".pickle",
     sig_figs: int = 8,
     dry_run: bool = True,
 ):
@@ -350,10 +353,10 @@ def round_sigfigs(num: float, sig_figs: int) -> float:
 
 
 def _remove_or_move_files(
-    fnames: List[str],
+    fnames: list[str],
     with_progress_bar: bool = True,
-    move_to: Optional[str] = None,
-    desc: Optional[str] = None,
+    move_to: str | None = None,
+    desc: str | None = None,
 ) -> None:
     """Remove files by filename.
 
@@ -387,11 +390,11 @@ def _remove_or_move_files(
 
 
 def load_parallel(
-    learners: List[adaptive.BaseLearner],
-    fnames: List[str],
+    learners: list[adaptive.BaseLearner],
+    fnames: list[str],
     *,
     with_progress_bar: bool = True,
-    max_workers: Optional[int] = None,
+    max_workers: int | None = None,
 ) -> None:
     r"""Load a sequence of learners in parallel.
 
@@ -420,8 +423,8 @@ def load_parallel(
 
 
 def save_parallel(
-    learners: List[adaptive.BaseLearner],
-    fnames: List[str],
+    learners: list[adaptive.BaseLearner],
+    fnames: list[str],
     *,
     with_progress_bar: bool = True,
 ) -> None:
@@ -479,7 +482,7 @@ def connect_to_ipyparallel(
     n: int,
     profile: str,
     timeout: int = 300,
-    folder: Optional[str] = None,
+    folder: str | None = None,
     client_kwargs=None,
 ):
     """Connect to an `ipcluster` on the cluster headnode.
@@ -512,7 +515,7 @@ def connect_to_ipyparallel(
     return client
 
 
-def _get_default_args(func: Callable) -> Dict[str, str]:
+def _get_default_args(func: Callable) -> dict[str, str]:
     signature = inspect.signature(func)
     return {
         k: v.default
@@ -528,7 +531,7 @@ def log_exception(log, msg, exception):
         log.exception(msg, exc_info=True)
 
 
-def maybe_lst(fname: Union[List[str], str]):
+def maybe_lst(fname: list[str] | str):
     if isinstance(fname, tuple):
         # TinyDB converts tuples to lists
         fname = list(fname)
@@ -561,11 +564,19 @@ class LRUCachedCallable(Callable[..., Any]):
     function : Callable[..., Any]
     max_size : int, optional
         Cache size of the LRU cache, by default 128.
+    with_cloudpickle : bool
+        Use cloudpickle for storing the data in memory.
     """
 
-    def __init__(self, function: Callable[..., Any], max_size: int = 128):
+    def __init__(
+        self,
+        function: Callable[..., Any],
+        max_size: int = 128,
+        with_cloudpickle: bool = False,
+    ):
         self.max_size = max_size
         self.function = function
+        self._with_cloudpickle = with_cloudpickle
         self._signature = signature(self.function)
         if max_size == 0:
             return
@@ -574,7 +585,7 @@ class LRUCachedCallable(Callable[..., Any]):
         self._cache_queue = manager.list()
         self._cache_lock = manager.Lock()
 
-    def _get_from_cache(self, key: str) -> Optional[Any]:
+    def _get_from_cache(self, key: str) -> Any | None:
         """Get a value from the cache by key."""
         if self.max_size == 0:
             value = None
@@ -585,14 +596,20 @@ class LRUCachedCallable(Callable[..., Any]):
                 self._cache_queue.append(key)
         if value is not None:
             found = True
-            value = value if value != _NONE_RETURN_STR else None
+            if value == _NONE_RETURN_STR:
+                value = None
+            elif self._with_cloudpickle:
+                value = cloudpickle.loads(value)
         else:
             found = False
         return found, value
 
     def _insert_into_cache(self, key: str, value: Any):
         """Insert a key value pair into the cache."""
-        value = value if value is not None else _NONE_RETURN_STR
+        if value is None:
+            value = _NONE_RETURN_STR
+        elif self._with_cloudpickle:
+            value = cloudpickle.dumps(value)
         with self._cache_lock:
             cache_size = len(self._cache_queue)
             self._cache_dict[key] = value
@@ -635,3 +652,164 @@ def shared_memory_cache(cache_size: int = 128):
         return functools.wraps(function)(LRUCachedCallable(function, cache_size))
 
     return cache_decorator
+
+
+def _prefix(fname: str | list[str] | tuple[str, ...]) -> str:
+    if isinstance(fname, (tuple, list)):
+        return f".{len(fname):08}_learners."
+    elif isinstance(fname, str):
+        return ".learner."
+    else:
+        raise TypeError("Incorrect type for fname.")
+
+
+def fname_to_learner_fname(fname: str | list[str] | tuple[str, ...]) -> str:
+    prefix = _prefix(fname)
+    if isinstance(fname, (tuple, list)):
+        fname = fname[0]
+    p = Path(fname)
+    return str(p.with_stem(f"{prefix}{p.stem}"))
+
+
+def fname_to_learner(fname: str | list[str] | tuple[str, ...]) -> adaptive.BaseLearner:
+    learner_name = fname_to_learner_fname(fname)
+    with open(learner_name, "rb") as f:
+        return cloudpickle.load(f)
+
+
+def _ensure_folder_exists(fnames: list[str | list[str] | tuple[str, ...]]) -> None:
+    if isinstance(fnames[0], (tuple, list)):
+        for _fnames in fnames:
+            _ensure_folder_exists(_fnames)
+    else:
+        folders = {Path(fname).parent for fname in fnames}
+        for folder in folders:
+            folder.mkdir(parents=True, exist_ok=True)
+
+
+def cloudpickle_learners(
+    learners,
+    fnames: list[str | list[str] | tuple[str, ...]],
+    with_progress_bar: bool = False,
+):
+    """Save a list of learners to disk using cloudpickle."""
+    _ensure_folder_exists(fnames)
+
+    for learner, fname in _progress(
+        zip(learners, fnames), with_progress_bar, desc="Cloudpickling learners"
+    ):
+        fname_learner = fname_to_learner_fname(fname)
+        with open(fname_learner, "wb") as f:
+            cloudpickle.dump(learner, f)
+
+
+def fname_to_dataframe(
+    fname: str | list[str] | tuple[str, ...], format: str = "parquet"
+) -> str | list[str]:
+    if format == "excel":
+        format = "xlsx"
+    if isinstance(fname, (tuple, list)):
+        fname = fname[0]
+    p = Path(fname)
+    return str(p.with_stem(f"dataframe.{p.stem}").with_suffix(f".{format}"))
+
+
+def save_dataframe(
+    fname: str | list[str] | tuple[str, ...],
+    format: _DATAFRAME_FORMATS = "parquet",
+    save_kwargs: dict[str, Any] | None = None,
+    expand_dicts: bool = True,
+    **to_dataframe_kwargs: Any,
+) -> Callable[[adaptive.BaseLearner], None]:
+    save_kwargs = save_kwargs or {}
+
+    def save(learner):
+        df = learner.to_dataframe(**to_dataframe_kwargs)
+        if expand_dicts:
+            df = expand_dict_columns(df)
+        fname_df = fname_to_dataframe(fname, format=format)
+        if format == "parquet":
+            df.to_parquet(fname_df, **save_kwargs)
+        elif format == "csv":
+            df.to_csv(fname_df, **save_kwargs)
+        elif format == "hdf":
+            if "key" not in save_kwargs:
+                save_kwargs["key"] = "data"
+            df.to_hdf(fname_df, **save_kwargs)
+        elif format == "pickle":
+            df.to_pickle(fname_df, **save_kwargs)
+        elif format == "feather":
+            df.to_feather(fname_df, **save_kwargs)
+        elif format == "excel":
+            df.to_excel(fname_df, **save_kwargs)
+        elif format == "json":
+            df.to_json(fname_df, **save_kwargs)
+        else:
+            raise ValueError(f"Unknown format {format}.")
+
+    return save
+
+
+_DATAFRAME_FORMATS = Literal[
+    "parquet", "csv", "hdf", "pickle", "feather", "excel", "json"
+]
+
+
+def expand_dict_columns(df: pd.DataFrame) -> pd.DataFrame:
+    if df.empty:
+        return df
+
+    for col, val in df.iloc[0].iteritems():
+        if isinstance(val, dict):
+            prefix = f"{col}."
+            x = pd.json_normalize(df.pop(col)).add_prefix(prefix)
+            x.index = df.index
+            for col in x:
+                assert col not in df, f"{col=} already exists in df."
+            df = df.join(x)
+    return df
+
+
+def load_dataframes(
+    fnames: list[str] | list[list[str]],
+    concat: bool = True,
+    read_kwargs: dict[str, Any] | None = None,
+    format: _DATAFRAME_FORMATS = "parquet",
+) -> pd.DataFrame | list[pd.DataFrame]:
+    read_kwargs = read_kwargs or {}
+    dfs = []
+    for fn in fnames:
+        fn_df = fname_to_dataframe(fn, format=format)
+        if not os.path.exists(fn):
+            continue
+        try:
+            if format == "parquet":
+                df = pd.read_parquet(fn_df, **read_kwargs)
+            elif format == "csv":
+                df = pd.read_csv(fn_df, **read_kwargs)
+            elif format == "hdf":
+                if "key" not in read_kwargs:
+                    read_kwargs["key"] = "data"
+                df = pd.read_hdf(fn_df, **read_kwargs)
+            elif format == "pickle":
+                df = pd.read_pickle(fn_df, **read_kwargs)
+            elif format == "feather":
+                df = pd.read_feather(fn_df, **read_kwargs)
+            elif format == "excel":
+                df = pd.read_excel(fn_df, **read_kwargs)
+            elif format == "json":
+                df = pd.read_json(fn_df, **read_kwargs)
+            else:
+                raise ValueError(f"Unknown format {format}.")
+        except Exception:
+            print(f"`{fn}`'s DataFrame ({fn_df}) could not be read.")
+            continue
+        df["fname"] = len(df) * [fn]
+        dfs.append(df)
+    if concat:
+        if dfs:
+            return pd.concat(dfs, axis=0)
+        else:
+            return pd.DataFrame()
+    else:
+        return dfs
